Require Import Vtac Program List Arith Even Div2 Relations.
Require Import basis countable lang subs model.
Require Import Classical FunctionalExtensionality.
Set Implicit Arguments.

(*----------------------------------------------------------------------------*)
(** ** Miscellaneous 'library' definitions and lemmas *)
(*----------------------------------------------------------------------------*)

Definition swap {A B} (x : A * B) := (snd x, fst x).

Lemma map_pair_fst_swap : 
  forall A B (f : A -> B) l,
    map_pair f (fun x => exp_of_val (fst x)) (map_pair id (swap (A:=val)) l) = 
    map_pair f (fun x => exp_of_val (snd x)) l. 
Proof. by induction l; intross; rewrite IHl. Qed.

Lemma map_pair_snd_swap : 
  forall A B (f : A -> B) l,
    map_pair f (fun x => exp_of_val (snd x)) (map_pair id (swap (B:=val)) l) = 
    map_pair f (fun x => exp_of_val (fst x)) l. 
Proof. by induction l; intross; rewrite IHl. Qed.

Lemma in_map_pair_swap :
  forall A B C x l, 
    In x (map_pair (A:= A) (C:=B*C) id swap l) <-> In (fst x, (swap (snd x))) l.
Proof.
  induction l as [|[? []]]; destruct x as [?[]]; simpls.
  split; intros; des; intuition vauto.
Qed.

Lemma list_assoc_in: forall A B eqdec l a b
    (FOUND: @list_assoc A B eqdec a l = Some b),
  In (a,b) l.
Proof.
induction l; intros; simpl; [discriminate|].
destruct a; simpl in *; destruct (eqdec a a0); subst; eauto.
inversion FOUND; subst; eauto.
Qed.

Lemma list_minus_emp_spec :
  forall A myeq l1 l2,
    (list_minus (A:=A) myeq l1 l2 = nil) <-> (forall x, In x l1 -> In x l2).
Proof.
  intros; induction l1 as [|a l1 IH]; simpl; intros; try tauto.
  generalize (list_mem_spec myeq a l2); destruct list_mem;
  intuition; congruence. 
Qed.

Lemma subst_closed_gg : 
  forall (f : val*val -> exp) A (gg : list (termvar*A*(val*val))) 
    (H: forall x, In x gg -> fv_exp (f (snd x)) = []),
  subst_closed (map_pair fst f gg).
Proof. induction gg; intross; auto. Qed.

Lemma eta3: forall A B C (r: A->B->C->Prop), r = fun a b c => r a b c.
Proof. by intros; repeat (eapply functional_extensionality; intros). Qed.

(*----------------------------------------------------------------------------*)
(** ** Helper lemmas for proving equalities up to proof irrelevance *)
(*----------------------------------------------------------------------------*)

Lemma sig_eq : 
  forall A (P: A -> Prop) (x x' : sig P), proj1_sig x = proj1_sig x' -> x = x'.
Proof. destruct x, x'; simpl; intro; subst; f_equal; apply proof_irr. Qed.

Lemma heap_eq : forall h1 h2, (forall x, heap_f h1 x = heap_f h2 x) -> h1 = h2.
Proof. 
  intros [] [] ?; simpl in *; apply functional_extensionality in H.
  subst; f_equal; apply proof_irr.
Qed. 

Lemma sref_eq :
  forall (s s': wst wref) (EQ: forall ty l l', `s ty l l' <-> `s' ty l l'), s = s'. 
Proof.
  by intros; eapply (@sig_eq _ _ s); eauto using @functional_extensionality, prop_ext.
Qed.

(*----------------------------------------------------------------------------*)
(** ** Properties of heaps *)
(*----------------------------------------------------------------------------*)

Program Definition hextm (h1 h2: heap) : heap :=
  {| heap_f x := match h1 x with None => h2 x | _ => h1 x end |}.
Next Obligation.
  destruct (heap_findom h1) as [d1 H1].
  destruct (heap_findom h2) as [d2 H2].
  exists (d1 ++ d2); intros.
  destruct (h1 x) as [] _eqn: X; clarify; eauto using in_or_app.
Qed.

Lemma hplus_det : 
  forall h1 h2 h (H: hplus h1 h2 h) h' (H': hplus h1 h2 h'), h' = h.
Proof. 
  intros; apply heap_eq; intro x.
  destruct (H x); destruct (H' x); des; congruence.
Qed.

Lemma hplus_hemp_l : forall h, hplus hemp h h.
Proof. vauto. Qed.

Lemma hplus_hemp_r : forall h, hplus h hemp h.
Proof. vauto. Qed.

Lemma hplus_hemp_lE : forall h h', hplus hemp h h' -> h = h'.
Proof.
  intros; apply heap_eq; intro x; destruct (H x); des; clarify; simpls; congruence.
Qed.

Lemma hplus_hemp_rE : forall h h', hplus h hemp h' -> h = h'.
Proof.
  intros; apply heap_eq; intro x; destruct (H x); des; clarify; simpls; congruence.
Qed.

Lemma hplus_hemp_inv : forall h h', hplus h h' hemp -> h = hemp /\ h' = hemp.
Proof.
  by intros; split; eapply heap_eq; intros x; specialize (H x); intuition.
Qed.

Lemma hplus_comm : forall h1 h2 h, hplus h1 h2 h -> hplus h2 h1 h.
Proof. by red; intros; edestruct H; vauto. Qed.

Lemma hplus_assoc_l : 
  forall h1 h2 h12 (R: hplus h1 h2 h12) h3 h (R': hplus h12 h3 h),
    exists h23, hplus h1 h23 h /\ hplus h2 h3 h23. 
Proof.
  intros; eexists (hextm h2 h3); split; intro x; simpl; 
    specialize (R x); specialize (R' x); destruct (h2 x); des; vauto;
    constructor (split; congruence). 
Qed.

Lemma hplus_assoc_r : 
  forall h2 h3 h23 (R: hplus h2 h3 h23) h1 h (R': hplus h1 h23 h),
    exists h12, hplus h1 h2 h12 /\ hplus h12 h3 h. 
Proof.
  intros; eexists (hextm h1 h2); split; intro x; simpl;
    specialize (R x); specialize (R' x); destruct (h1 x); des; vauto;
    constructor (split; congruence). 
Qed.

Lemma hplus_none :
  forall h1 h2 h (P: hplus h1 h2 h) loc (N: h loc = None),
    h1 loc = None /\ h2 loc = None.
Proof.
  intros; specialize (P loc); simpls; des; split; congruence.
Qed.

Lemma hplus_some1 :
  forall h1 h2 h (P: hplus h1 h2 h) loc v (N: h1 loc = Some v),
    h loc = Some v.
Proof. intros; specialize (P loc); simpls; des; congruence. Qed.

Lemma hplus_some2 :
  forall h1 h2 h (P: hplus h1 h2 h) loc v (N: h2 loc = Some v),
    h loc = Some v.
Proof. intros; specialize (P loc); simpls; des; congruence. Qed.

Lemma hplus_inl : forall h1 h2 h12 (HP : hplus h1 h2 h12) l v (EH : h1 l = Some v),
  h12 l = Some v.
Proof.
  by intros; specialize (HP l); des; [rewrite HP0 in EH | rewrite HP].
Qed.

Lemma hplus_inr : forall h1 h2 h12 (HP : hplus h1 h2 h12) l v (EH : h2 l = Some v),
  h12 l = Some v.
Proof.
  by intros; specialize (HP l); des; [rewrite HP | rewrite HP0 in EH].
Qed.

Lemma hplus_assoc : forall h1 h2 h3 h12 h23 h123
    (HP12 : hplus h1 h2 h12)
    (HP123 : hplus h12 h3 h123)
    (HP23 : hplus h2 h3 h23),
  hplus h1 h23 h123.
Proof.
  intros; assert (HP := hplus_assoc_l HP12 HP123); des.
  by rewrite (hplus_det HP0 HP23).
Qed.

Lemma hplus_disjoint : forall h1 h2 h12 (HP : hplus h1 h2 h12) l
    v1 (IN1 : h1 l = Some v1) v2 (IN2 : h2 l = Some v2),
  False.
Proof.
  by intros; destruct (HP l); des; [rewrite H0 in IN1 | rewrite H0 in IN2].
Qed.

Definition hunion' (h1 h2 : heap) l :=
  match h1 l with Some v1 => Some v1 | _ => h2 l end.

Definition hunion (h1 h2 : heap) : heap.
  refine (Build_heap (hunion' h1 h2) _).
  intros.
  destruct (heap_findom h1) as (dom1, ?); destruct (heap_findom h2) as (dom2, ?).
  exists (app dom1 dom2); intros l ?.
  by unfold hunion'; case_eq (h1 l); intros; apply in_or_app; [left | right]; eauto.
Defined.

Lemma hunion_inl : forall (h1 h2 : heap) l v (INL : h1 l = Some v),
  hunion h1 h2 l = Some v.
Proof.
  by intros; simpl; unfold hunion'; rewrite INL.
Qed.

Lemma hunion_inr : forall (h1 h2 : heap) l v (INR : h2 l = Some v) h12 (DJ : hplus h1 h2 h12),
  hunion h1 h2 l = Some v.
Proof.
  intros; simpl; unfold hunion'; specialize (DJ l); revert DJ.
  by rewrite INR; case_eq (h1 l); intros; des; vauto.
Qed.

Definition hrestrict' (h : heap) (dom : location -> Prop) l :=
  match LEM (dom l) with
  | left _ => h l
  | right _ => None
  end.

Definition hrestrict (h  : heap )(dom : location -> Prop) : heap.
  refine (Build_heap (hrestrict' h dom) _).
  destruct (heap_findom h) as (hdom, ?); exists hdom; unfold hrestrict'.
  by intro l; case_eq (LEM (dom l)); intros; eauto.
Defined.


Hint Immediate hplus_hemp_l hplus_hemp_r hplus_comm. 

(*----------------------------------------------------------------------------*)
(** ** Properties of the underlying untyped language *)
(*----------------------------------------------------------------------------*)

Lemma is_value_of_expE :
  forall e, is_value_of_exp e -> { v | e = exp_of_val v }.
Proof.
  induction e; intross; des.
  - by exists (Vvar x).
  - by exists (Vunit).
  - by exists (Vint n).
  - by exists (Vtrue).
  - by exists (Vfalse).
  - eby destruct IHe1; simpls; destruct IHe2; simpls; eexists (Vpair _ _); simpls; f_equal.
  - eby destruct IHe; simpls; eexists (Vinl _); simpls; f_equal.
  - eby destruct IHe; simpls; eexists (Vinr _); simpls; f_equal.
  - eby eexists (Vloc _); simpls; f_equal.
  - eby destruct IHe; simpls; eexists (Vroll _); simpls; f_equal.
  - by exists (Vfix f x e).
  - by exists (Vgen e).
  - by destruct IHe as [v ?]; simpls; exists (Vpack v); simpls; f_equal.
Qed.

Lemma is_value_of_val : forall v, is_value_of_exp (exp_of_val v).
Proof. induction v; intros; simpl in *; auto. Qed.
Hint Resolve is_value_of_val. 

(** Values cannot reduce. *)

Lemma not_reduce_val :
   forall h (v : val) h' e', reduce h v h' e' -> False.
Proof.
  intros; remember (exp_of_val v) as e; revert v Heqe.
  induction H; try by intros []. 
  intros.
  cut (exists v : val, e = v); [by intro; des; eauto|].
  rename v into vv; revert vv Heqe; clear.
  destruct KK; simpls; intros; eauto;
    try (by destruct vv; simpls; clarify; eauto).
Qed.

Lemma not_reduce_val2 :
  forall h e h' e', reduce h e h' e' -> is_value_of_exp e -> False.
Proof.
  intros; destruct (is_value_of_expE _ H0); clarify.
  eapply not_reduce_val, H.
Qed.

Lemma exp_of_val_inj :
  forall v v', exp_of_val v = exp_of_val v' -> v = v'.
Proof.
  induction v; destruct v'; intross; clarify; try by rewrite (IHv _ H).
  by rewrite (IHv1 _ H1), (IHv2 _ H).
Qed.

(** The operational semantics is deterministic. *)

Lemma reduce_det :
   forall h e h1 e1 (R1: reduce h e h1 e1) h2 e2 (R2: reduce h e h2 e2),
     h1 = h2 /\ e1 = e2.
Proof.
  induction 1; inversion 1; clarify;
  try (destruct KK; simpls; clarify);
  try (destruct KK0; simpls; clarify; try destruct (IHR1 _ _ RED); clarify);
  try (by destruct (not_reduce_val2 RED)); 
  try (by destruct (not_reduce_val2 R1));
  repeat match goal with 
    H : exp_of_val _ = exp_of_val _ |- _ => apply exp_of_val_inj in H; clarify
  end.
  split; try done.
  apply heap_eq; intro x; generalize (f_equal (fun f => f x) H0); simpl.
    by destruct Peano_dec.eq_nat_dec.
Qed.

Lemma reduce_appctx_scont_det:
  forall h K e h' e' h'' e''
    (RC: reduce h (appctx_scont_exp K e) h' e')
    (RED: reduce h e h'' e''),
   e' = appctx_scont_exp K e''.
Proof. intros; apply (reduce_det RC (Rsctx K RED)). Qed.

Lemma reduce_app_cont_det:
  forall h K e h' e' h'' e''
    (RC: reduce h (app_cont K e) h' e')
    (RED: reduce h e h'' e''),
   e' = app_cont K e''.
Proof. intros; apply (reduce_det RC (Rctx RED K)). Qed.

Lemma appctx_scont_exp_val : forall K (e : exp) (v : val) (NV : forall v' : val, e <> v')
  (E : appctx_scont_exp K e = v), False.
Proof.
  by intros; destruct K; simpls; destruct v;
    try (by auto); try (by inversion E; eapply NV; eauto).
Qed.

Lemma reduces_val_det :
   forall h e h1 e1 (R1: reduces h e h1 e1) (v1 : val) (E1 : e1 = v1)
              h2 e2 (R2: reduces h e h2 e2) (v2 : val) (E2 : e2 = v2),
     h1 = h2 /\ v1 = v2.
Proof.
  intros ? ? ? ? ?; induction R1; intros; subst; subst.
    destruct R2; subst; eauto using exp_of_val_inj.
    by exfalso; eapply not_reduce_val; eauto.
  destruct R2; subst.
    by exfalso; eapply not_reduce_val; eauto.
  subst; edestruct reduce_det; [by apply STEP | by apply STEP0 |].
  by subst; eapply IHR1; eauto.
Qed.

Lemma reduces_stuck_det :
   forall h e h1 e1 (R1: reduces h e h1 e1) (NRS1 : forall h' e', ~ reduce h1 e1 h' e')
                    (NV1 : ~ is_value_of_exp e1)
              h2 e2 (R2: reduces h e h2 e2) (NRS2 : forall h' e', ~ reduce h2 e2 h' e')
                    (NV2 : ~ is_value_of_exp e2),
     h1 = h2 /\ e1 = e2.
Proof.
  intros ? ? ? ? ?; induction R1; intros.
    by destruct R2; subst; auto; exfalso; eapply NRS1; eauto.
  destruct R2; subst. 
    by exfalso; eapply NRS2; eauto.
  edestruct reduce_det; [by apply STEP | by apply STEP0 |].
  by subst; eapply IHR1; eauto.
Qed.

Lemma not_diverges_converges :
  forall h e, ~ diverges h e -> converges h e.
Proof.
  red; intros; apply NNPP; intro X.
  cut (forall h' e', reduces h e h' e' -> exists h'' e'', reduce h' e' h'' e'').
    clear X; intro Z; destruct H.
    revert h e Z; cofix X; intros.
    destruct (classic (exists h' e', reduce h e h' e')); des.
      by econstructor; [edone|eapply X]; eauto using reduces_step.
    by specialize (Z _ _ (@reduces_nil _ _ _ _ (refl_equal _) (refl_equal _))).
  intros.
  eapply not_ex_all_not, not_ex_all_not, not_and_or in X; destruct X as [X|X].
    eby destruct X.
  do 2 (eapply not_all_ex_not in X; des).
  eby eexists; eexists; eapply NNPP.
Qed.

Lemma not_reduce_app_int : forall h e n (v : val) (E : e = Eapp (Vint n) v) h' e',
  ~ reduce h e h' e'.
Proof.
  intros; intro R; induction R; inversion E; destruct KK; inversion H0.
    by eapply not_reduce_val; rewrite H1 in R; fold (exp_of_val (Vint n)) in R; eauto.
  by rewrite H2 in R; eapply not_reduce_val; eauto.
Qed.

Lemma not_reduce_inst_int : forall h e n (E : e = Einst (Vint n)) h' e',
  ~ reduce h e h' e'.
Proof.
  intros; intro R; induction R; inversion E; destruct KK; inversion H0.
  by rewrite H1 in R; fold (exp_of_val (Vint n)) in R; eapply not_reduce_val; eauto.
Qed.

Lemma not_value_app : forall e1 e2 (V : is_value_of_exp (Eapp e1 e2)), False.
Proof.
  by intros; inversion V.
Qed.
  
Lemma not_value_inst : forall e (V : is_value_of_exp (Einst e)), False.
Proof.
  by intros; inversion V.
Qed.
  
Lemma not_value_cont : forall e (NV : ~ is_value_of_exp e)
    K (V : is_value_of_exp (app_cont K e)), False.
Proof.
  by intros; induction K; simpls; des; apply IHK.
Qed.

Lemma reduces_trans : 
  forall h e h' e' (R: reduces h e h' e') 
    h'' e'' (R': reduces h' e' h'' e''),
    reduces h e h'' e''.
Proof. induction 1; subst; eauto using reduces_step. Qed.

Lemma reduces_diverges : 
  forall h e h' e' (R: reduces h e h' e') (D: diverges h' e'),
    diverges h e.
Proof. induction 1; subst; eauto using diverges_step. Qed.

Lemma diverges_reduces : 
  forall h e h' e' (R: reduces h e h' e') (D: diverges h e),
    diverges h' e'.
Proof. 
  induction 1; subst; intross; destruct D.
  eapply (reduce_det STEP) in STEP0; des; clarify; eauto.
Qed.

Lemma reduces_cont : 
  forall h e h' e' (D: reduces h e h' e') K, 
    reduces h (app_cont K e) h' (app_cont K e').
Proof. induction 1; subst; econstructor (by eauto using Rctx). Qed.

Lemma diverges_cont : 
  forall h e (D: diverges h e) K, diverges h (app_cont K e).
Proof. cofix X; intros; destruct D; econstructor; eauto using Rctx. Qed.

Ltac dc :=
  clarify; repeat match goal with
  | H: reduce ?x ?y _ _ , H' : reduce ?x ?y _ _ |- _ =>
    destruct (reduce_det H H'); clear H'; clarify
  | H: hplus hemp _ _ |- _ => apply hplus_hemp_lE in H; clarify
  | H: hplus _ hemp _ |- _ => apply hplus_hemp_rE in H; clarify
  | H: hplus ?x ?y _, H' : hplus ?x ?y _ |- _ =>
    apply (hplus_det H) in H'; clarify
  end.

Lemma diverges_no_stuck: 
  forall h e, diverges h e -> ~ gets_stuck h e.
Proof.
  unfold gets_stuck. red; intros; des; red in STUCK.
  induction RED; inv H; dc; eauto.
Qed.

Lemma reduces_val_no_stuck: 
  forall h e h' (v : val), reduces h e h' v -> ~ gets_stuck h e.
Proof.
  unfold gets_stuck; red; intros; des; red in STUCK.
  induction RED; inv H; dc; eauto using not_reduce_val.
Qed.

Lemma reduces_val_no_diverges :
 forall h e h' (v : val), reduces h e h' v -> ~ diverges h e.
Proof.
  red; intros; eapply diverges_reduces in H; eauto.
  eby inv H; edestruct not_reduce_val.
Qed.

Lemma not_is_value_app_cont : forall e (NV : ~ is_value_of_exp e) K,
  ~ is_value_of_exp (app_cont K e).
Proof.
  by intros; induction K; simpl; intro V; try destruct V; contradiction.
Qed.

Lemma app_cont_eq : forall e1 e2 (NV1 : ~ is_value_of_exp e1) (NV2 : ~ is_value_of_exp e2)
    (ATOM1 : forall e1' (NV1' : ~ is_value_of_exp e1') K1' (E1 : app_cont K1' e1' = e1),
               K1' = Khole)
    (ATOM2 : forall e2' (NV2' : ~ is_value_of_exp e2') K2' (E1 : app_cont K2' e2' = e2),
               K2' = Khole)
    K1 K2 (E : app_cont K1 e1 = app_cont K2 e2),
  K1 = K2 /\ e1 = e2.
Proof.
  induction K1; simpl; induction K2; simpl; intros; try inversion E;
    try match goal with
        | _:_ |- Khole = ?K /\ _ =>
          by assert (FF : K = Khole); [eapply ATOM1; try apply NV2; eauto | inversion FF]
        | _:_ |- ?K = Khole /\ _ =>
          by assert (FF : K = Khole); [eapply ATOM2; try apply NV1; eauto | inversion FF]
    end;
    try (by edestruct IHK1 as [->]; auto; try (apply exp_of_val_inj in H0; by rewrite H0));
    try (by exfalso; absurd (is_value_of_exp v); auto; rewrite <- H0;
      apply not_is_value_app_cont);
    try (by exfalso; absurd (is_value_of_exp v); auto; rewrite H0;
      apply not_is_value_app_cont).
Qed.

Lemma atom_app_val_val : forall e (NV : ~ is_value_of_exp e)
    K (v1 v2 : val) (E : app_cont K e = Eapp v1 v2),
  K = Khole.
Proof.
  by induction K; intros; simpls;
     inversion E; eapply not_is_value_app_cont with (K:=K) in NV;
     try rewrite H0 in NV; try rewrite H1 in NV; induction v1.
Qed.

Lemma atom_inst_val : forall e (NV : ~ is_value_of_exp e)
    K (v : val) (E : app_cont K e = Einst v),
  K = Khole.
Proof.
  by induction K; intros; simpls;
    inversion E; eapply not_is_value_app_cont with (K:=K) in NV; 
    try rewrite H0 in NV; try rewrite H1 in NV; induction v.
Qed.

Lemma reduce_app_cont_inv: forall K h e h' e' 
    (R : reduce h (app_cont K e) h' e') 
    (EXP: ~ is_value_of_exp e),
  exists e'', reduce h e h' e''.
Proof.
  intros K h e; rewrite <- appctx_scont_of_cont.
  setoid_rewrite <- rev_involutive; setoid_rewrite fold_left_rev_right.
  generalize (rev (list_scont_of_cont K)) as Ks; clear K.
  intros Ks; revert h e; induction Ks; eauto.
  intros; simpls; exploit IHKs; eauto.
    by red; intros X; apply EXP; destruct a; eauto; inv X.
  by intros RED; des; destruct a; dependent destruction RED; 
      try (by contradiction EXP; eauto); 
      destruct KK; inv x; eauto;
      exfalso; eauto using not_reduce_val.
Qed.

Lemma not_reduce_cont : forall K h e h' e'
    (NR : forall h' e', ~ reduce h e h' e') 
    (NV : ~ is_value_of_exp e)
    (R : reduce h (app_cont K e) h' e'), 
  False.
Proof. by intros; red in NR; edestruct reduce_app_cont_inv; eauto. Qed.

(*----------------------------------------------------------------------------*)
(** ** Properties of the typing judgments *)
(*----------------------------------------------------------------------------*)

Lemma hastype_wfenv :
  forall D G e ty (HT: hastype D G e ty), wfenv D G.
Proof. induction 1; eauto; inv H. Qed.

Lemma hastype_wftyp :
  forall D G e ty (HT: hastype D G e ty), wftyp D G ty.
Proof.
  induction 1; eauto; vauto;
  repeat (rewrite ?list_minus_app in *; des; match goal with
    | H : _ ++ _ = nil |- _ => eapply app_eq_nil in H
    | H : _ |- _ => rewrite H; []; simpl
    | H : list_assoc ?myeq ?x ?l = _ |- _ =>
      pose proof (list_assoc_spec myeq x l); destruct (list_assoc myeq x l); clarify
    | H : wftyp _ _ _ |- _ => inversion_clear H; simpl in *; clarify; eauto
    | |- wftyp _ _ _ => constructor; simpl; eauto using lc_vsubst, lc_vsubstD
    | H : wfenv _ _ |- _ => inversion_clear H; simpl in *; clarify; eauto
    | |- list_minus _ _ _ = [] => apply list_minus_emp_spec; intros; eauto
    | H : list_minus _ _ _ = [] |- In _ _ =>
      solve [eapply list_minus_emp_spec; [eapply H|eauto using in_ftn_vsubst_mon]]
    | H : In _ (ftn_typ (vsubst _ _ _)) |- _ => apply in_ftn_vsubst in H; simpl in H; destruct H; eauto
  end).
Qed.

Lemma wfctx_wfenv :
  forall C D G ty D' G' ty' (HT: wfctx C D G ty D' G' ty'), wfenv D G.
Proof. induction 1; eauto; inv H. Qed.

Lemma wfctx_wftyp :
  forall C D G ty D' G' ty' (HT: wfctx C D G ty D' G' ty'), wftyp D' G' ty'.
Proof.
  induction 1; eauto; vauto;
  repeat (rewrite ?list_minus_app in *; des; match goal with
    | H : _ ++ _ = nil |- _ => eapply app_eq_nil in H
    | H : _ |- _ => rewrite H; []; simpl
    | H : list_assoc ?myeq ?x ?l = _ |- _ =>
      pose proof (list_assoc_spec myeq x l); destruct (list_assoc myeq x l); clarify
    | H : wftyp _ _ _ |- _ => inversion_clear H; simpl in *; clarify; eauto
    | |- wftyp _ _ _ => constructor; simpl; eauto using lc_vsubst, lc_vsubstD
    | H : wfenv _ _ |- _ => inversion_clear H; simpl in *; clarify; eauto
    | |- list_minus _ _ _ = [] => apply list_minus_emp_spec; intros; eauto
    | H : list_minus _ _ _ = [] |- In _ _ =>
      solve [eapply list_minus_emp_spec; [eapply H|eauto using in_ftn_vsubst_mon]]
    | H : In _ (ftn_typ (vsubst _ _ _)) |- _ => apply in_ftn_vsubst in H; simpl in H; destruct H; eauto
    | H : phastype _ _ _ _ |- _ => apply hastype_erase, hastype_wftyp in H; simpl in H; eauto
  end).
Qed.

(*----------------------------------------------------------------------------*)
(** * Language-independent lemmas about worlds *)
(*----------------------------------------------------------------------------*)

Program Definition wid : world_local := 
  wsingle (fun _ _ _ _ => False) (fun _ h h' => h = hemp /\ h' = hemp) _ _ _ _ _.

Lemma wstable_id : wstable wid.
Proof. repeat red; eauto. Qed.

Lemma winhabited_id : winhabited (wlift wid).
Proof. exists (sref0,()); repeat exists hemp; eby repeat split; eauto. Qed.

Lemma wconsistent_id : wconsistent (wlift wid).
Proof. red; inversion 2; intros; simpls; des; clarify. Qed.

Hint Resolve  wstable_id winhabited_id wconsistent_id.

Lemma wstuck_mon: forall rf rv rf' rv' ty e1 e2 
    (WS: wstuck rf rv ty e1 e2) (LEF: rf <3= rf') (LEV: rv <3= rv'),
  wstuck rf' rv' ty e1 e2.
Proof. by intros; inv WS; [econstructor 1|econstructor 2]; eauto using vclos_mon. Qed.

Lemma minvr_glob: forall (w : world), wglob _ (minvr w).
Proof.
  split; [|split]; repeat intro.
  + induction V; eapply wval_closed; eauto.
  + induction R; eauto using wval_mon.
  + split; [|split]; unnw; intros; try destruct sp; try destruct R; eauto using wval_mon.
Qed.

Lemma minvr_init: forall w r (G: wglob _ r), minvr w <4= r.
Proof.
  intros; rdes G; rhdes GSUP; induction PR; eauto using wval_mon.
Qed.

Lemma sref0_hemp: forall s r (HR: wheap wref () s r hemp hemp), s = sref0.
Proof.
  intros; apply sref_eq; intuition; repeat red in HR; des.
  by specialize (HSAT _ _ _ H); des.
Qed.

(*----------------------------------------------------------------------------*)
(** ** World disjointness *)
(*----------------------------------------------------------------------------*)

Definition wdisj (w1 w2 : world_local) := ndisj (wnam w1) (wnam w2).

Lemma wdisj_comm : 
  forall w1 w2, wdisj w1 w2 -> wdisj w2 w1.
Proof. eby red; red; intros; eapply H. Qed.

Lemma wdisj_prod_l : 
  forall w1 w2 w (D: wdisj w1 w) (D': wdisj w2 w), 
  wdisj (wprod w1 w2) w.
Proof. by intros; intros n []; [apply D | apply D']. Qed.

Lemma wdisj_prod_r : 
  forall w1 w2 w (D: wdisj w w1) (D': wdisj w w2),
  wdisj w (wprod w1 w2).
Proof. intros; auto using wdisj_prod_l, wdisj_comm. Qed.

Hint Resolve wdisj_prod_l wdisj_prod_r.
Hint Immediate wdisj_comm.

(*----------------------------------------------------------------------------*)
(** ** Basic properties of worlds *)
(*----------------------------------------------------------------------------*)

Lemma wexp_val :
  forall w g (gOK: wglob w g) s0 s (PUB: wextpub w s s0) ty v v' 
    (V: forall (gOK: wglob w g), vclos (g s) ty v v'),
    wexp w g s0 s ty v v'. 
Proof. econstructor 2; eauto using reduces_nil. Qed.

Lemma wexp_core_mon :
  forall w r r' g g' s0' s0 s ty h hf hh e h' hf' hh' e'
    (WE: wexp_core w r g s0' s ty h hf hh e h' hf' hh' e') 
    (LTR: r <4= r') (LTG: g <4= g') (PUB: wextpub w s0' s0), 
  wexp_core w r' g' s0 s ty h hf hh e h' hf' hh' e'. 
Proof.
  intros until s0; revert r r' g g'.
  cofix X; intros; destruct WE; vauto.
    by econstructor 2; eauto using wextpub_trans, vclos_mon, wheap_mon. 
  econstructor 3; eauto using wheap_mon.
    by destruct EREL; econstructor (by eauto using vclos_mon).
  do 2 red in KREL |- *; intros; eapply X; try eapply KREL with (g' := g'0); eauto.
Qed.

Lemma wexp_mon :
  forall w g s0' s0 s ty e1 e2 (WE: wexp w g s0' s ty e1 e2) (PUB: wextpub w s0' s0), 
  wexp w g s0 s ty e1 e2. 
Proof. by repeat red; intros; eauto using wexp_core_mon. Qed.

Lemma wexp_core_reduces: forall h e h1 e1 h' e' h2 e2 w s' s s0 r g ty
   hf hh (PLUS : hplus h hf hh)
   hf' hh' (PLUS' : hplus h' hf' hh')
   hh1 (PLUS1 : hplus h1 hf hh1)
   hh2 (PLUS2 : hplus h2 hf' hh2)
   (RED : reduces hh e hh1 e1)
   (RED' : reduces hh' e' hh2 e2)
   (EXT : wext w s' s)
   (R: wexp_core w r g s0 s' ty h1 hf hh1 e1 h2 hf' hh2 e2),
   wexp_core w r g s0 s ty h hf hh e h' hf' hh' e'.
Proof.
  destruct 8; dc; eauto using 
    wexp_Diverge, wexp_Terminate, wexp_Call,
    reduces_diverges, reduces_trans, wext_trans.
Qed.

Lemma wstuck_prod:
  forall (w1 : world_local) w2 sp s r r' ty e1 e2
    (EREL : wstuck (wval (wprod w1 w2) sp s r) r' ty e1 e2),
  wstuck (wval w1 sp (fst s) r) r' ty e1 e2 \/ wstuck (wval w2 sp (snd s) r) r' ty e1 e2.
Proof. intros; inv EREL; inv FREL; vauto. Qed.

Lemma wstuck_lift:
  forall w sp s r r' ty e1 e2,
    wstuck (wval (wlift w) sp s r) r' ty e1 e2 <->
    wstuck (wval w (fst s) (snd s) r) r' ty e1 e2.
Proof. by split; intros; destruct H; simpls; des; try discriminate; vauto. Qed.

Lemma wstuck_disj:
  forall r r1 r2
    (gEQ : forall (ty : flextyp) v v',
      r ty v v' <-> r1 ty v v' \/ r2 ty v v')
    g ty e1 e2
    (EREL : wstuck r g ty e1 e2),
  wstuck r1 g ty e1 e2 \/ wstuck r2 g ty e1 e2.
Proof.
  intros; destruct EREL; clarify; eapply gEQ in FREL; des; vauto. 
Qed.
Implicit Arguments wstuck_disj [r r1 r2 g ty e1 e2].

Lemma wglob_prod1: forall 
    w1 w2 (WD: wdisj w1 w2)
    g (gOK: wglob (wlift (wprod w1 w2)) g) s2, 
  wglob (wlift w1) (fun s => g (fst s,(snd s, s2))).
Proof.
  intros; rdes gOK; intros; split; [|split]; unnw; intros; simpls; des; eauto.

  assert (X:=GSUP (fst s, (snd s,s2))); rdes X.
  repeat split; simpls; unnw; intros.
  - by apply LE; intuition. 
  - apply GEREF in R; simpls.
    assert (Z:= wlocal_nonref w2 (fst s) s2 (g (fst s,(snd s,s2))) ty v v').
    by intuition. 
  - rdes NM.
    + by eapply GENAM in R; left; des; clarify; elimtype False; eauto using wnam_supports.
    + eapply GENAM in R; right; des; clarify; red; eauto. 
      by elimtype False; eauto using wnam_supports.
Qed.

Lemma wglob_prod2: forall 
    w1 w2 (WD: wdisj w1 w2)
    g (gOK: wglob (wlift (wprod w1 w2)) g) s1, 
    wglob (wlift w2) (fun s => g (fst s,(s1, snd s))).
Proof.
  intros; rdes gOK; intros; split; [|split]; unnw; intros; simpls; des; eauto.

  assert (X:= GSUP (fst s, (s1, snd s))); rdes X.
  repeat split; simpls; unnw; intros.
  - by apply LE; intuition. 
  - apply GEREF in R; simpls.
    assert (Z:= wlocal_nonref w1 (fst s) s1 (g (fst s,(s1, snd s))) ty v v').
    by intuition. 
  - rdes NM.
    + by eapply GENAM in R; left; des; clarify; elimtype False; eauto using wnam_supports.
    + eapply GENAM in R; right; des; clarify; red; eauto.
      by elimtype False; eauto using wnam_supports.
Qed.

(*----------------------------------------------------------------------------*)
(** ** Context application *)
(*----------------------------------------------------------------------------*)

Lemma wexp_core_app_cont :
  forall w g (gOK: wglob w g) s s0 s0' ty ty' K1 K2
    (KREL: forall s' (EXTP: wextpub w s' s0') (EXT: wext w s' s) 
           g' (g'OK: wglob w g') (LT : g <4= g'),
           wcont w g' s0 s' ty' ty K1 K2)
    r (LT: g <4= r) 
    h h' (HREL: wheap w () s (g s) h h')
    hf hh (HPLUS: hplus h hf hh)
    hf' hh' (HPLUS': hplus h' hf' hh')
    e e' (EREL: wexp_core w r g s0' s ty' h hf hh e h' hf' hh' e'),
  wexp_core w r g s0 s ty h hf hh (app_cont K1 e) h' hf' hh' (app_cont K2 e').
Proof.
  cofix X; intros; destruct EREL.
  - by clear X; eauto using wexp_Diverge, diverges_cont.
  - by clear X; exploit KREL; 
       eauto using wexp_core_mon, wexp_core_reduces, reduces_cont, wext_trans, wextpub_in_wext.
  - econstructor 3; try eapply EREL; try eapply HREL; try edone.
    + instantiate (2 := cont_of_list_scont (list_scont_of_cont K1 ++ 
                                            list_scont_of_cont K0)).
      instantiate (1 := cont_of_list_scont (list_scont_of_cont K2 ++ 
                                            list_scont_of_cont K3)).
      do 2 red in KREL0 |- *; intros.
      rewrite !app_cont_of_list_scont, !fold_right_app, 
           <- !app_cont_of_list_scont, !cont_of_list_scont_of_cont. 
      by eapply X; clear X; intros; try eapply KREL; eauto using wext_trans; try done.
    + rewrite app_cont_of_list_scont, fold_right_app, 
        <- !app_cont_of_list_scont, !cont_of_list_scont_of_cont. 
      by apply reduces_cont.
    + rewrite app_cont_of_list_scont, fold_right_app, 
        <- !app_cont_of_list_scont, !cont_of_list_scont_of_cont. 
      by apply reduces_cont.
Qed.

Lemma wexp_app_cont :
  forall w s0 s0' g (gOK: wglob w g) s ty ty' K1 K2
    (KREL: forall s' (EXTP: wextpub w s' s0') (EXT: wext w s' s) 
           g' (g'OK: wglob w g') (LT : g <4= g'),
           wcont w g' s0 s' ty' ty K1 K2) 
    e e' (EREL: wexp w g s0' s ty' e e'),
    wexp w g s0 s ty (app_cont K1 e) (app_cont K2 e').
Proof.
  by do 2 red; intros; eapply wexp_core_app_cont; eauto.
Qed.

(*----------------------------------------------------------------------------*)
(** ** External calls *)
(*----------------------------------------------------------------------------*)

(** Weird induction principle for external call lemma *)

Inductive term hh e : Prop :=
  | term_nil : 
    forall (NORED: forall hha ea, ~ reduce hh e hha ea),
      term hh e
  | term_step : 
    forall (STRED: forall hhn en  (RED: reduces hh e hhn en) 
                 hhn' en' (RED': reduce hhn en hhn' en'), term hhn' en'),
    term hh e. 

Lemma reduces_end : 
  forall h e h' e' (S: reduce h e h' e') 
    h'' e'' (R: reduces h' e' h'' e''), 
    exists h1 e1, reduces h e h1 e1 /\ reduce h1 e1 h'' e''.
Proof.
  intros; revert h e S; induction R; subst; intros; eauto using reduces_nil.
  destruct (IHR _ _ STEP); des; eauto using reduces_step.
Qed.

Lemma reduces2D : 
  forall h e h' e' (S: reduces h e h' e') 
    h'' e'' (R: reduces h e h'' e''), 
    (exists h1 e1, reduce h' e' h1 e1 /\ reduces h1 e1 h'' e'') \/ reduces h'' e'' h' e'. 
Proof.
  induction 1; destruct 1; vauto; dc; eauto.
Qed.

Lemma reduces_term : forall h' e' h e (R: reduces h e h' e')  (T: term h' e') , term h e.
Proof.
  induction 1; intros; vauto.
  specialize (IHR T).
  econstructor 2; intros.
  inv RED; dc.
  destruct IHR; eauto. 
  eby inv R; [inv RED0|]; edestruct NORED.
Qed.

Lemma reduces_term2 : forall h e h' e' (R: reduces h e h' e')  (T: term h e) , term h' e'.
Proof.
  induction 1; intros; vauto.
  eapply (reduces_end STEP) in R; des.
  destruct T; eauto. 
  eby edestruct NORED.
Qed.

Lemma not_diverges_term :
  forall h e, ~ diverges h e -> term h e.
Proof.
  intros; eapply not_diverges_converges in H; destruct H; des.
  eapply reduces_term; try eassumption; vauto.
Qed.

Lemma wexp_core_external1:
  forall w (CONS: wconsistent w) g (gOK: wglob w g) r 
    (gEQ: forall s ty v v', g s ty v v' 
             <-> wval w () s (g s) ty v v' \/ r s ty v v') 
    hh e (T : term hh e)
    h hf (PLUS: hplus h hf hh)
    h' hf' hh' (PLUS': hplus h' hf' hh')
    ty s0 s e' (EE : wexp_core w g g s0 s ty h hf hh e h' hf' hh' e'),
  wexp_core w r g s0 s ty h hf hh e h' hf' hh' e'.
Proof.
  induction 4.

  Case 0. (* base case *)
  destruct 3; vauto.
  case (wstuck_disj (gEQ s') EREL); clear EREL; intro EREL; [|by vauto]; dc.
  apply CONS in EREL; eauto; destruct EREL as (e1' & e2' & H1 & H2 & HH);
   specialize (H1 hh1); specialize (H2 hh2); specialize (HH _ _ HREL _ _ PLUS1 _ _ PLUS2).
  inv RED; try eby edestruct NORED.
  eby edestruct NORED; eapply Rctx.

  Case 1. (* inductive step *)
  destruct 3; vauto.
  case (wstuck_disj (gEQ s') EREL); clear EREL; intro EREL; [|by vauto]; dc.
  apply CONS in EREL; eauto; destruct EREL as (e1' & e2' & H1 & H2 & HH);
    specialize (H1 hh1); specialize (H2 hh2); specialize (HH _ _ HREL _ _ PLUS1 _ _ PLUS2).

  eapply wexp_core_reduces; try eapply RED; try eapply RED'; eauto using wext_trans.
  eapply wexp_core_reduces; try eapply reduces_step, reduces_nil; 
    try eapply Rctx; try eapply H1; try eapply H2; eauto using wext_trans.
  by eapply (H _ _ RED _ _ (Rctx H1 _)), wexp_core_app_cont, HH;
     try red; intros; eauto using wext_trans.
Qed.

Lemma wexp_core_external2:
  forall w (CONS: wconsistent w) g (gOK: wglob w g) r 
    (gEQ: forall s ty v v', g s ty v v' 
             <-> wval w () s (g s) ty v v' \/ r s ty v v') 
    hh' e' (T : term hh' e')
    h hf hh (PLUS: hplus h hf hh)
    h' hf' (PLUS': hplus h' hf' hh')
    ty s0 s e (EE : wexp_core w g g s0 s ty h hf hh e h' hf' hh' e'),
  wexp_core w r g s0 s ty h hf hh e h' hf' hh' e'.
Proof.
  induction 4.

  Case 0. (* base case *)
  destruct 3; vauto.
  case (wstuck_disj (gEQ s') EREL); clear EREL; intro EREL; [|by vauto]; dc.
  apply CONS in EREL; eauto; destruct EREL as (e1' & e2' & H1 & H2 & HH);
   specialize (H1 hh1); specialize (H2 hh2); specialize (HH _ _ HREL _ _ PLUS1 _ _ PLUS2).
  inv RED'; try eby edestruct NORED.
  eby edestruct NORED; eapply Rctx.

  Case 1. (* inductive step *)
  destruct 3; vauto.
  case (wstuck_disj (gEQ s') EREL); clear EREL; intro EREL; [|by vauto]; dc.
  apply CONS in EREL; eauto; destruct EREL as (e1' & e2' & H1 & H2 & HH);
   specialize (H1 hh1); specialize (H2 hh2); specialize (HH _ _ HREL _ _ PLUS1 _ _ PLUS2).

  eapply wexp_core_reduces; try eapply RED; try eapply RED'; eauto using wext_trans.
  eapply wexp_core_reduces; try eapply reduces_step, reduces_nil; 
    try eapply Rctx; try eapply H1; try eapply H2; eauto using wext_trans.
  by eapply (H _ _ RED' _ _ (Rctx H2 _)), wexp_core_app_cont, HH; 
     try red; intros; eauto using wext_trans.
Qed.

Lemma wexp_core_external :
  forall w (CONS: wconsistent w) g (gOK: wglob w g) r 
    (gEQ: forall s ty v v', g s ty v v' 
             <-> wval w () s (g s) ty v v' \/ r s ty v v') 
    h hf hh (PLUS: hplus h hf hh)
    h' hf' hh' (PLUS': hplus h' hf' hh')
    ty e e' s0 s (H: wexp_core w g g s0 s ty h hf hh e h' hf' hh' e'),
    wexp_core w r g s0 s ty h hf hh e h' hf' hh' e'. 
Proof.
  intros.
  destruct (classic (diverges hh e)) as [|NDIV].
    destruct (classic (diverges hh' e')) as [|NDIV]; [by vauto|].
    by eauto using wexp_core_external2, not_diverges_term.
  by eauto using wexp_core_external1, not_diverges_term.
Qed.

Lemma wexpgen_external :
  forall w (CONS: wconsistent w) g (gOK: wglob w g) r 
    (gEQ: forall s ty v v', g s ty v v' 
             <-> wval w () s (g s) ty v v' \/ r s ty v v') 
    s0 s ty e e' (H: wexp w g s0 s ty e e'),
  wexpgen w r g s0 s ty e e'.
Proof. by repeat intro; eapply wexp_core_external; eauto. Qed.

Corollary wexp_no_external :
  forall w (CONS: wconsistent w) 
    g (gOK: wglob w g)  
    (gEQ: forall s ty v v', g s ty v v' <-> wval w () s (g s) ty v v')
    ty e e' s0 s (H: wexp w g s0 s ty e e'),
  wexpgen w (fun _ _ _ _ => False) g s0 s ty e e'.
Proof.
  intros; eapply wexpgen_external; eauto.
  intros; specialize (gEQ s1 ty0 v v'); intuition.
Qed.

Lemma wrefst_inj1 : forall (s : wst wref) l1
    ty l2 (IN : `s ty l1 l2)
    ty' l2' (IN' : `s ty' l1 l2'),
  ty = ty' /\ l2 = l2'.
Proof.
  intros; destruct s; simpls.
  by rrdes w; edestruct w0 as (A, ?); [apply IN | apply IN' | specialize (A eq_refl)].
Qed.

Lemma wrefst_inj2 : forall (s : wst wref) l2
    ty l1 (IN : `s ty l1 l2)
    ty' l1' (IN' : `s ty' l1' l2),
  ty = ty' /\ l1 = l1'.
Proof.
  intros; destruct s; simpls.
  by rrdes w; edestruct w0 as (?, A); [apply IN | apply IN' | specialize (A eq_refl)].
Qed.

Implicit Arguments wrefst_inj1 [s l1 ty l2 ty' l2'].
Implicit Arguments wrefst_inj2 [s l2 ty l1 ty' l1'].

