Require Import basis lang subs countable.
Require Import Vtac Program List Relations.

Set Implicit Arguments.

(*----------------------------------------------------------------------------*)
(** * Basic library definitions *)
(*----------------------------------------------------------------------------*)

(* begin hide *)
Obligation Tactic := 
  repeat intro; 
  try (eby solve [unnw; program_simpl]); 
  try (eby simpls; des; des; unnw; clarify; eauto 10).
(* end hide *)

(*----------------------------------------------------------------------------*)
(** ** Divergence and stuckness *)
(*----------------------------------------------------------------------------*)

Inductive reduces h e (h': heap) (e': exp) : Prop :=
| reduces_nil (HEQ: h' = h) (EEQ: e' = e)
| reduces_step h'' e'' 
    (STEP: reduce h e h'' e'')
    (RED: reduces h'' e'' h' e'). 
 
CoInductive diverges h e : Prop :=
| diverges_step : forall h' e' 
    (STEP: reduce h e h' e')
    (DIV: diverges h' e'), 
  diverges h e.

Definition converges h e : Prop :=
  exists h' e', 
       <<RED: reduces h e h' e'>>
    /\ <<STUCK: forall h'' e'', ~ reduce h' e' h'' e''>>.

Definition gets_stuck h e : Prop :=
  exists h' e',
     << RED: reduces h e h' e' >>
  /\ << STUCK : forall h'' e'', ~ reduce h' e' h'' e'' >>
  /\ << NOEXP: ~ is_value_of_exp e' >>.

(*----------------------------------------------------------------------------*)
(** * Contextual equivalence *)
(*----------------------------------------------------------------------------*)

Definition ctx_equiv D G ty p1 p2 :=
  forall c ty' (WFC: wfctx c D G ty nil nil ty') h, 
    <<DIV: diverges h (app_ctx c p1)>> <-> <<DIV': diverges h (app_ctx c p2)>>.

(*----------------------------------------------------------------------------*)
(** * Various relations *)
(*----------------------------------------------------------------------------*)

Implicit Types (v : val) (e: exp) (h : heap).

(** Flexible types are those for which value equivalence is not fixed. *)

Inductive flextyp : Type := 
  | FTarrow (ty1 ty2 : typ)
  | FTref (ty : typ)
  | FTforall (ty : typ)
  | FTname (a : typename).

Definition unflex (ty : flextyp) : typ :=
  match ty with
  | FTarrow ty1 ty2 => Tarrow ty1 ty2
  | FTref ty' => Tref ty'
  | FTforall ty' => Tforall ty'
  | FTname a => Tname a
  end.
Coercion unflex : flextyp >-> typ.

Lemma unflex_inj : forall t1 t2, unflex t1 = unflex t2 -> t1 = t2.
Proof.
  by destruct t1; intros; destruct t2; inversion H; try rewrite <- H1, <- H2.
Qed.


(** Here are the basic domains that we will be using: *)

Definition vrelf := flextyp -> val -> val -> Prop.
Definition vrel := typ -> val -> val -> Prop.
Definition erel := typ -> exp -> exp -> Prop.
Definition hrel := heap -> heap -> Prop.

Hint Unfold vrelf vrel erel hrel.

(*----------------------------------------------------------------------------*)
(** * Value Closure *)
(*----------------------------------------------------------------------------*)

(** The following definition, [vclos], lifts the definition of value 
    equivalence from flexible types only to all types. *)

Inductive vclos_gen (r: vrelf) (vr: vrel) : vrel :=
  | vclos_gen_name : forall aa v v' (VR: r (FTname aa) v v'), vclos_gen r vr (Tname aa) v v'
  | vclos_gen_unit : vclos_gen r vr Tunit Vunit Vunit
  | vclos_gen_int  : forall (n : nat), vclos_gen r vr Tint (Vint n) (Vint n)
  | vclos_gen_bool_true : vclos_gen r vr Tbool Vtrue Vtrue
  | vclos_gen_bool_false : vclos_gen r vr Tbool Vfalse Vfalse
  | vclos_gen_prod:
    forall ty1 v1 v1' (VR1: vr ty1 v1 v1') 
           ty2 v2 v2' (VR2: vr ty2 v2 v2'),
      vclos_gen r vr (Tprod ty1 ty2) (Vpair v1 v2) (Vpair v1' v2')
  | vclos_gen_sum_inl:
    forall ty1 v1 v1' (VR: vr ty1 v1 v1') ty2 (CT : lc_typ 0 ty2),
      vclos_gen r vr (Tsum ty1 ty2) (Vinl v1) (Vinl v1')
  | vclos_gen_sum_inr:
    forall ty2 v2 v2' (VR: vr ty2 v2 v2') ty1 (CT : lc_typ 0 ty1),
      vclos_gen r vr (Tsum ty1 ty2) (Vinr v2) (Vinr v2')
  | vclos_gen_arrow:
    forall ty1 ty2 v v' (VR: r (FTarrow ty1 ty2) v v'),
      vclos_gen r vr (Tarrow ty1 ty2) v v'
  | vclos_gen_ref:
    forall ty v v' (VR: r (FTref ty) v v'),
      vclos_gen r vr (Tref ty) v v'
  | vclos_gen_mu:
    forall ty v v' (VR: vr (vsubst 0 (Tmu ty) ty) v v'),
      vclos_gen r vr (Tmu ty) (Vroll v) (Vroll v')
  | vclos_gen_forall:
    forall ty v v' (VR: r (FTforall ty) v v'),
      vclos_gen r vr (Tforall ty) v v'
  | vclos_gen_exists:
    forall ty' (WFty': lc_typ 0 ty') ty v v' (VR: vr (vsubst 0 ty' ty) v v'),
      vclos_gen r vr (Texists ty) (Vpack v) (Vpack v').
Hint Constructors vclos_gen.

Inductive vclos (r : vrelf) (ty: typ) (v1 v2: val) : Prop :=
| vclos_intro (REL: vclos_gen r (vclos r) ty v1 v2).
Hint Constructors vclos.

Lemma vclos_gen_mon : forall r0 r ty v1 v2 (REL: vclos_gen r0 r ty v1 v2) r' (LE: r <3= r'),
  vclos_gen r0 r' ty v1 v2.
Proof. intros; dependent destruction ty; inversion REL; eauto. Qed.

Lemma vclos_induction : 
  forall 
    r ty v1 v2 (EL: vclos r ty v1 v2)
    P (PF: forall ty v1 v2 (IH: vclos_gen r P ty v1 v2), P ty v1 v2),
  P ty v1 v2.
Proof.
  fix 5.
  intros; apply PF; destruct EL; destruct REL;
  econstructor; eauto; eapply vclos_induction; eauto.
Qed.

Lemma vclos_induction_strong : 
  forall 
    r ty v1 v2 (EL: vclos r ty v1 v2)
    P (PF: forall ty v1 v2 
             (IH: vclos_gen r (fun ty' v1' v2' => P ty' v1' v2' /\ vclos r ty' v1' v2') ty v1 v2), 
           P ty v1 v2),
  P ty v1 v2.
Proof.
  intros; cut (P ty v1 v2 /\ vclos r ty v1 v2); [intuition|].
  revert P PF; eapply (vclos_induction EL); clear ty v1 v2 EL; intros.
  split; eauto using vclos_gen_mon.
  econstructor; eapply vclos_gen_mon; eauto.
  by simpl; intros; hdes; eauto.
Qed.

Definition vrelf_closed (r : flextyp -> val -> val -> Prop) := 
  forall ty v v' (V: r ty v v'),
    <<LC: lc_typ 0 ty>> /\ <<CV: fv_exp v = [] >> /\ <<CV': fv_exp v' = [] >>.

Hint Unfold vrelf_closed.

Lemma vclos_mon: 
  forall g g' ty v v' (VC: vclos g ty v v') (LT: g <3= g'), 
    vclos g' ty v v'.
Proof. 
  intros until v'; intros REL; eapply (vclos_induction REL); intros. 
  by destruct IH; simpls; econstructor; eauto.
Qed.

Lemma vclos_closed : 
  forall r ty v v' (VREL : vclos r ty v v') (C : vrelf_closed r), 
    <<LC: lc_typ 0 ty>> /\ <<CV: fv_exp v = [] >> /\ <<CV': fv_exp v' = [] >>.
Proof.
  intros; eapply (vclos_induction VREL); intros. 
  destruct IH; simpl; subst; try specialize (C _ _ _ VR); des;
    simpls; repeat (match goal with [H: _ = [] |- _] => rewrite H end);
    eauto 10 using lc_vsubstD, ftn_vsubst_mon_closed.
Qed.

(*----------------------------------------------------------------------------*)
(** * Dependent World *)
(*----------------------------------------------------------------------------*)

Definition funval (f: val) : Prop := 
  forall v, exists e, forall h, reduce h (Eapp f v) h e.

Definition tyfunval (f: val) : Prop :=
  exists e, forall h, reduce h (Einst f) h e.

Structure world_dep (pst : Type) (pext : relation pst) := 
  { wnam : typename -> Prop 
  ; wst  : Type
  ; wext : relation wst
  ; wextpub : relation wst
  ; wval  : pst -> wst -> vrelf -> vrelf
  ; wheap : pst -> wst -> vrelf -> hrel
  ; wext_trans      : forall x y (EXT: wext x y) z (EXT': wext y z), wext x z 
  ; wextpub_refl    : forall x, wextpub x x 
  ; wextpub_trans   : forall x y (EXT: wextpub x y) z (EXT': wextpub y z), wextpub x z 
  ; wextpub_in_wext : forall x y (EXTP: wextpub x y), wext x y
  ; wval_mon        : forall sp s r ty v v' (WV: wval sp s r ty v v')
                        sp' (RP: pext sp' sp) 
                        s' (R: wext s' s) 
                        r' (LT: r <3= r'),
                      wval sp' s' r' ty v v'
  ; wval_closed     : forall sp s r (C: vrelf_closed r), vrelf_closed (wval sp s r)
  ; wheap_mon       : forall sp s r h h' (PR: wheap sp s r h h') r' (LT: r <3= r'),
                        wheap sp s r' h h'
  ; wnam_supports   : forall sp s r a v v' (WV: wval sp s r (FTname a) v v'), wnam a 
  ; wval_arrow_wf   : forall sp s r ty' ty f1 f2
                             (L : wval sp s r (FTarrow ty' ty) f1 f2),
                        <<FV1: funval f1 >> /\ <<FV2: funval f2 >>
  ; wval_forall_wf  : forall sp s r ty f1 f2
                             (L : wval sp s r (FTforall ty) f1 f2),
                        <<FV1: tyfunval f1 >> /\ <<FV2: tyfunval f2 >>
  }.

Implicit Arguments wval_mon [w sp sp' s s' r r' ty v v'].
Implicit Arguments wval_closed [w sp s r ty v v'].
Hint Immediate wextpub_refl wextpub_in_wext.

Lemma wext_refl pst (pext : relation pst) (w : world_dep pext) x : wext w x x.
Proof. apply wextpub_in_wext, wextpub_refl. Qed.
Hint Immediate wext_refl.

(*----------------------------------------------------------------------------*)
(** * Full World *)
(*----------------------------------------------------------------------------*)

Definition world := world_dep (fun x y: unit => True).

Inductive minvr (w : world) (s: wst w) (ty: flextyp) (v1 v2: val) : Prop := 
| minvr_intro r (REL: wval w () s r ty v1 v2) (MON: r <3= (minvr w s)).
Hint Constructors minvr.

(*----------------------------------------------------------------------------*)
(** * Ref World *)
(*----------------------------------------------------------------------------*)

(** We define a ref world that defines the meaning of the
    mutable reference type. *)

Definition wrefst (s : typ -> location -> location -> Prop) :=
     (forall ty loc1 loc2 (IN : s ty loc1 loc2), <<LC: lc_typ 0 ty >>)
  /\ (forall ty loc1 loc2 (IN : s ty loc1 loc2) ty' loc1' loc2' (IN' : s ty' loc1' loc2'),
        (forall (EQ: loc1 = loc1'), <<TEQ: ty = ty' >> /\ <<LEQ: loc2 = loc2' >>) /\
        (forall (EQ: loc2 = loc2'), <<TEQ: ty = ty' >> /\ <<LEQ: loc1 = loc1' >>))
  /\ (exists tydom ldom, 
       forall ty l1 l2 (IN: s ty l1 l2), 
         <<INTY: In ty tydom>> /\ <<INL1: In l1 ldom>> /\ <<INL2: In l2 ldom>>).

Program Definition wref : world :=
  {| wnam n := False
   ; wst := { s : typ -> location -> location -> Prop | wrefst s } 
   ; wext s s' := s' <3= s
   ; wextpub s s' := s' <3= s
   ; wval u s r ty v v' := exists loc, exists loc', exists ty',
                              <<TEQ: ty = FTref ty'>> 
                           /\ <<VEQ: v = Vloc loc >> 
                           /\ <<VEQ': v' = Vloc loc'>> 
                           /\ <<TLL: s ty' loc loc'>>
   ; wheap u s r h h' := 
       <<HSAT : forall ty loc loc' (R: s ty loc loc'), 
                exists v v', 
                <<HL: h loc = Some v>> /\ <<HL': h' loc' = Some v'>> /\
                <<VC: vclos r ty v v'>> >> /\
       <<HDOM1 : forall l v (IN: h l = Some v), exists ty l', s ty l l' >> /\
       <<HDOM2 : forall l' v' (IN: h' l' = Some v'), exists ty l, s ty l l' >> |}.
Next Obligation. 
des; subst; destruct s; simpl; rrdes w; eauto 15.
Qed.
Next Obligation.
des; unnw; split; eauto; intros; eapply HSAT in R; des; eauto 7 using vclos_mon.
Qed.

Program Definition sref0 : wst wref := exist _ (fun _ _ _ => False) _.
Next Obligation.
red; repeat (split; intros; vauto).
Existential 1 := nil. Existential 1 := nil.
Qed.

(*----------------------------------------------------------------------------*)
(** * Local World *)
(*----------------------------------------------------------------------------*)

Structure world_local := 
  { wldlocal :> world_dep (wext wref)
  ; wlocal_nonref : forall sp s r ty v v', wval wldlocal sp s r (FTref ty) v v' -> False
  }.

(*----------------------------------------------------------------------------*)
(** * Product of two local worlds *)
(*----------------------------------------------------------------------------*)

Program Definition wprod (w1 w2 : world_local) : world_local :=
  {| wldlocal := 
  {| wnam := nunion (wnam w1) (wnam w2)
   ; wext s s' := <<WE: wext w1 (fst s) (fst s')>> /\ <<WE': wext w2 (snd s) (snd s')>>
   ; wextpub s s' := <<WEP: wextpub w1 (fst s) (fst s')>> /\ <<WEP': wextpub w2 (snd s) (snd s')>>
   ; wval sp s r ty v v' := wval w1 sp (fst s) r ty v v' \/ wval w2 sp (snd s) r ty v v'
   ; wheap sp s r h h' := exists h1 h1' h2 h2', 
                          <<WH:  wheap w1 sp (fst s) r h1 h1'>> /\
                          <<WH': wheap w2 sp (snd s) r h2 h2'>> /\
                          <<HP: hplus h1 h2 h>> /\ <<HP': hplus h1' h2' h'>> |} |}.
Next Obligation. simpls; des; eauto 10 using wext_trans, wextpub_trans. Qed. 
Next Obligation. simpls; des; eauto 10 using wext_trans, wextpub_trans. Qed.
Next Obligation. simpls; des; eauto 10 using wval_mon. Qed.
Next Obligation. by des; eapply (wval_closed _ _ C V). Qed.
Next Obligation. eby des; repeat eexists; try eapply wheap_mon. Qed.
Next Obligation. by des; constructor (eby eapply wnam_supports). Qed.
Next Obligation. simpls; des; eauto using wval_arrow_wf. Qed.
Next Obligation. simpls; des; eauto using wval_forall_wf. Qed.
Next Obligation. simpls; des; eauto using wlocal_nonref. Qed.

(*----------------------------------------------------------------------------*)
(** * Lifting of a local world *)
(*----------------------------------------------------------------------------*)

Program Definition wlift (w : world_local) : world :=
{| wnam := nunion (wnam wref) (wnam w)
 ; wext s s' := <<WRE: wext wref (fst s) (fst s')>> /\ <<WRE': wext w (snd s) (snd s')>>
 ; wextpub s s' := <<WREP: wextpub wref (fst s) (fst s')>> /\ <<WREP': wextpub w (snd s) (snd s')>>
 ; wval sp s r ty v v' := wval wref () (fst s) r ty v v' \/ wval w (fst s) (snd s) r ty v v'
 ; wheap sp s r h h' := exists h1 h1' h2 h2', 
                        <<WH: wheap wref () (fst s) r h1 h1'>> /\
                        <<WH': wheap w (fst s) (snd s) r h2 h2'>> /\
                        <<HP: hplus h1 h2 h>> /\ <<HP': hplus h1' h2' h'>> |}.
Next Obligation. des; eauto 10 using wext_trans. Qed.
Next Obligation. des; eauto 10 using wextpub_trans. Qed.
Next Obligation. des; eauto 10 using wval_mon. Qed.
Next Obligation. by des; apply (wval_closed _ _ C V). Qed.
Next Obligation. 
des; simpls; do 4 eexists; repeat split; try eapply wheap_mon; des; unnw; eauto.
by intros; specialize (HSAT _ _ _ R); des; eauto 10 using vclos_mon.
Qed.
Next Obligation. by des; constructor (eby eapply wnam_supports). Qed.
Next Obligation. by des; eauto using wval_arrow_wf. Qed.
Next Obligation. by des; eauto using wval_forall_wf. Qed.

(*----------------------------------------------------------------------------*)
(** * Single State Worlds *)
(*----------------------------------------------------------------------------*)

Program Definition wsingle (VR: vrelf->vrelf ) (HR: vrelf->hrel) 
  (VRMON: forall r r' (LE: r <3= r'), VR r <3= VR r')
  (HRMON: forall r r' (LE: r <3= r'), HR r <2= HR r')
  (VRCLOS: forall r (C: vrelf_closed r), vrelf_closed (VR r))
  (WFARR: forall r ty' ty f1 f2 (L: VR r (FTarrow ty' ty) f1 f2), 
            <<FV: funval f1>> /\ <<FV': funval f2>>)
  (WFINST: forall r ty f1 f2 (L: VR r (FTforall ty) f1 f2),
             <<FV: tyfunval f1>> /\ <<FV': tyfunval f2>>)
  : world_local 
  :=
  {| wldlocal :=
  {| wnam := fun _ => False
   ; wst := unit
   ; wext := fun _ _ => True
   ; wextpub := fun _ _ => True
   ; wval := fun _ _ r ty v v' => 
     match ty with | FTref _ => False | FTname _ => False | _ => VR r ty v v' end
   ; wheap := fun _ _ => HR |} |}.
Next Obligation. destruct ty; simpls; eapply VRMON; eauto. Qed.
Next Obligation. destruct ty; eapply VRCLOS; eauto; contradiction. Qed.

(*----------------------------------------------------------------------------*)
(** * Global Knowledge *)
(*----------------------------------------------------------------------------*)

(** A custom order between two value relations that requires the
    meaning of reference and abstract types to agree between the
    two relations. *)

Definition ltfun N (r r' : vrelf) :=
  <<LE: forall ty v v' (R: r ty v v'), r' ty v v' >> /\
  <<GEREF: forall ty v v' (R: r' (FTref ty) v v'), r (FTref ty) v v' >> /\
  <<GENAM: forall a v v' (NM: N a) (R: r' (FTname a) v v'), r (FTname a) v v' >>.

(** A global value relation [g] is consistent w.r.t. a world [w] if
    it is (i) well-formed; 
        (ii) monotonic w.r.t state transitions 
    and (iii) relates any two values that are related by [w]. *)

Definition wglob (w : world) (g : wst w -> vrelf) :=
  <<GCLO: forall s, vrelf_closed (g s) >> /\
  <<GMON: forall s s' ty v v' (R: g s ty v v') (EXT: wext w s' s), g s' ty v v' >> /\
  <<GSUP: forall s, ltfun (wnam w) (wval w () s (g s)) (g s) >>.

(*----------------------------------------------------------------------------*)
(** * Expression and Continuation Equivalence *)
(*----------------------------------------------------------------------------*)

(** 
   Co-inductive definition of when two (closed) expressions are related. 
*)

Inductive wstuck (Rf : vrelf) (Rv : vrelf) ty e1 e2 : Prop :=
  | wstuck_App
      ty' (f1 v1 f2 v2 : val)
      (EQ1  : e1 = Eapp f1 v1)
      (EQ2  : e2 = Eapp f2 v2)
      (FREL : Rf (FTarrow ty' ty) f1 f2)
      (VREL : vclos Rv ty' v1 v2)
  | wstuck_Inst
      ty' ty0 (f1 f2 : val)
      (EQ1  : e1 = Einst f1)
      (EQ2  : e2 = Einst f2)
      (EQ3  : ty = vsubst 0 ty' ty0)
      (LC   : lc_typ 0 ty')
      (FREL : Rf (FTforall ty0) f1 f2).
Hint Constructors wstuck.

Definition wexp_u wexp_core (w: world) (r g: wst w -> vrelf) (s0 s: wst w) (ty: typ) e e' :=
  forall 
    h h' (HREL: wheap w () s (g s) h h')
    hf hh (HPLUS: hplus h hf hh)
    hf' hh' (HPLUS': hplus h' hf' hh'),
    wexp_core w r g s0 s ty h hf hh e h' hf' hh' e' : Prop.

Definition wcont_u wexp_core w g s0 s ty' ty k k' :=
  forall v v' (VREL: vclos (g s) ty' v v'),
  wexp_u wexp_core w g g s0 s ty (app_cont k v) (app_cont k' v').

CoInductive wexp_core (w : world) r g s0 s ty h hf hh e h' hf' hh' e' : Prop :=
| wexp_Diverge
    (DIV: diverges hh e) 
    (DIV': diverges hh' e')

| wexp_Terminate
     s' (EXT: wext w s' s) (EXTP: wextpub w s' s0)
     (v1 v2 : val) (VREL: vclos (g s') ty v1 v2)
     h1 h2 (HREL: wheap w () s' (g s') h1 h2)
     hh1 (PLUS1: hplus h1 hf hh1)
     hh2 (PLUS2: hplus h2 hf' hh2)
     (RED: reduces hh e hh1 v1)
     (RED': reduces hh' e' hh2 v2)

| wexp_Call
     s' (EXT: wext w s' s)
     h1 h2 (HREL: wheap w () s' (g s') h1 h2)
     e1 e2 ty' (EREL: wstuck (r s') (g s') ty' e1 e2)
     K1 K2 (KREL: forall s'' (EXTP: wextpub w s'' s') g' (g'OK: wglob w g') (LT : g <4= g'),
                  wcont_u wexp_core w g' s0 s'' ty' ty K1 K2)
     hh1 (PLUS1: hplus h1 hf hh1)
     hh2 (PLUS2: hplus h2 hf' hh2)
     (RED: reduces hh e hh1 (app_cont K1 e1))
     (RED': reduces hh' e' hh2 (app_cont K2 e2))
.

Definition wexp w g := wexp_u wexp_core w g g.

Definition wexpgen w r g := wexp_u wexp_core w r g.

(** Two continuations are related if whenever we apply them to 
    related arguments, we get related expressions. *)

Definition wcont := wcont_u wexp_core.

(*----------------------------------------------------------------------------*)
(** * Program Equivalence *)
(*----------------------------------------------------------------------------*)

Definition wstable (w : world_local) :=
  forall 
    g (gOK: wglob (wlift w) g)
    sp s h1 h2 (HR: wheap w sp s (g (sp,s)) h1 h2)
    sp' (EXT: wext _ sp' sp)
    hs1 hs2 (SHR: wheap wref () sp' (g (sp',s)) hs1 hs2)
    hh1 (DIS: hplus hs1 h1 hh1)
    hh2 (DIS: hplus hs2 h2 hh2),
  exists s', <<WEP: wextpub _ s' s>> /\ <<WH: wheap _ sp' s' (g (sp',s')) h1 h2>>.

(** We say that a world is inhabited iff it relates the empty heap to itself. *)

Definition winhabited (w : world) := 
  forall g (gOK: wglob w g), exists s0, wheap w () s0 (g s0) hemp hemp.

(** A world is consistent if it any functions it declares as 
    being equivalent are indeed so when applied to equivalent
    arguments. *)

Definition wconsistent (w : world) :=
  forall
    g (gOK: wglob w g)
    s ty e1 e2 (APP: wstuck (wval w () s (g s)) (g s) ty e1 e2),
    exists e1', exists e2',
      <<BETA:  forall h, reduce h e1 h e1' >> /\
      <<BETA': forall h, reduce h e2 h e2' >> /\
      <<WEXP: wexp w g s s ty e1' e2'>>.

(** Two expressions are related by a world [w] in a typing environment [D; G],
    if they are related by [w] after substituting any pairwise related values
    for the expressions' free variables. *)

Inductive wequiv D G w ty e1 e2 : Prop :=
  wequiv_intro
    (envOK: forall x ty0 (IN: In (x,ty0) G), lc_typ 0 ty0)
    (tyOK: lc_typ 0 ty)
    (FV1 : list_minus eq_termvar (fv_exp e1) (map fst G) = [])
    (FV2 : list_minus eq_termvar (fv_exp e2) (map fst G) = [])
    (WEXP: forall s 
             g (gOK: wglob w g)
             d (DEQ: map fst d = D)
               (dOK: forall x (IN: In x d), lc_typ 0 (snd x))
             gg (GEQ: map fst gg = G) 
                (ggOK: forall x ty0 v1 v2 (IN: In ((x,ty0),(v1,v2)) gg), 
                       vclos (g s) (tsubst_typ d ty0) v1 v2),
           wexp w g s s (tsubst_typ d ty) 
             (subst_exp (map_pair fst (fun x => exp_of_val (fst x)) gg) e1)
             (subst_exp (map_pair fst (fun x => exp_of_val (snd x)) gg) e2)).

Hint Constructors wequiv.

Definition weq D G w ty e1 e2 : Prop :=
  <<INH : winhabited w>> /\
  <<CS : wconsistent w>> /\
  <<EQ : wequiv D G w ty e1 e2>>.

Hint Unfold weq.

(** Finally, two expressions are equivalent at a given type, if there 
    exists an inhabited, non-ref world, whose product with the global
    [wref] world is consistent and relates the two expressions. *)

Definition wreq (D : list typename) (G: list (termvar * typ)) ty e1 e2 :=
  forall N (INF: countably_infinite N), 
  exists w : world_local,
    <<NAM: wnam w <1= N >> /\
    <<WSTA: wstable w >> /\
    <<WEQ: weq D G (wlift w) ty e1 e2>>.
